package neuroet;

import drasys.or.matrix.*;
import drasys.or.linear.algebra.*;
import drasys.or.cont.*;
import drasys.or.*;

/*The abstract trainer class needs to be created with an output layer and
one call to train with the answers[] will do one iteration of training
*/


abstract class Trainer extends Object
{
	protected OutputLayer ol;

	Trainer(OutputLayer o)
	{
		ol = o;
	}

	abstract public double Train(double[] answers) throws TrainingException;


        private final static String[] trainerNames = {"Conjugate Gradient", "Levenberg-Marquardt", "Standard Backprop"};

        public static String[] GetTrainerNames()
        {
          return (String[])trainerNames.clone();
        }

	public static Trainer GetTrainer(int trainer, OutputLayer ol)
	{
			switch (trainer)
			{
                                case 0:
                                        return (new ConjugateGradientTrainer(ol));

                                case 1:
					return (new LMTrainer(ol));

				case 2:
					return (new StandardBackpropTrainer(ol));

				default:
					throw new IllegalArgumentException ("Invalid Trainer Specified:" + trainer);
			}


	}

}

/*On-line, cyclic implementation*/
class StandardBackpropTrainer extends Trainer
{
	StandardBackpropTrainer (OutputLayer o)
	{
		super(o);
	}

	public double Train(double[] answers) throws TrainingException
	{


		double[] currError = new double[1];

		for (int x = 0; x < answers.length; x++)
		{
			currError[0] = answers[x] - ol.GetOutput();
			ol.AdjustWeightsAndBiases(currError);
		}


		return ArrayUtilities.SumOfSquares(NeuralNet.RunNetwork(ol, answers));

	}




}

class LMTrainer extends Trainer
{

	private double mu;
	static final private double v = 10.0;
	static final private Algebra alg = new Algebra();


	LMTrainer (OutputLayer o)
	{
		super(o);
		mu = 0.01;

	}


	public double Train(double[] answers) throws TrainingException
	{

		double[] currError = new double[1];
		double[][] jacobian = new double[answers.length][];

		ContiguousMatrix jacobianMatrix;
		DenseMatrix tj;
		DenseVector latter;
		DenseMatrix result;

		double[] mods = null;

		double currentSS;
		double[] es = new double[answers.length];
		double[] diagonal;
		double error = Double.POSITIVE_INFINITY;

			try{

			for (int x = 0; x < es.length; x++)
			{
				es[x] = currError[0] = answers[x] - ol.GetOutput();
				jacobian[x] = ol.CalculateJacobianRow(currError);
			}


			currentSS = ArrayUtilities.SumOfSquares(es);

			jacobianMatrix = new ColumnMajorMatrix(jacobian);

			tj = alg.transpose(jacobianMatrix);

			latter = alg.multiply(tj, new DenseVector(es));//J^t * v(x)

			jacobianMatrix = alg.multiply(tj, jacobianMatrix);// = J^t * j

			diagonal = jacobianMatrix.diagonalVector().getArray();



			do
			{

				mu *= v;

                                if (mu < 0.1E-50)
                                  mu = 0.1E-25;

				//System.err.println(mu);//debugging

				if (mu > 1.0E16)
				{
					throw new TrainingException ("Can't Advance any further? MU = " + mu);
				}




				jacobianMatrix.setDiagonal(new DenseVector(ArrayUtilities.IncrementArray(diagonal, mu))); //should be J^t * J + I



				try{//could optimize by moving outwards

					result = alg.invert(jacobianMatrix);

				}catch (AlgebraException a){System.out.println("Couldn't invert");continue;} //if we can't invert, increase mu


				mods = alg.multiply(result, latter).getArray();
				ArrayUtilities.NegateArray(mods);
				error = NeuralNet.RunWithModsAndAverageError(ol, answers, mods);



			}while (error > currentSS);





			ol.ModifyNetwork(mods);

			mu /= v * v;  //since we always multiply by v earlier


			return error;
			}catch (AlgebraException r){r.printStackTrace(); throw new TrainingException("Couldn't Train: " + r);}

        }





}





class ConjugateGradientTrainer extends Trainer
{

	private double beta;
	private double[] previousDirection;
        private double[] previousGradient;
	private int n;
	private int l;
	private double lastRanError;
        private static final double goldenSection = 1.0 - 0.618;
        private static final double resolution = 0.001;

	ConjugateGradientTrainer (OutputLayer o)
	{
		super (o);
		beta = 0.0;
		previousDirection = new double[o.NumParams()];//just needs the correct length
                previousGradient = new double[previousDirection.length];
		n = previousDirection.length; //number of variables we have
		l = 1;
		lastRanError = Double.POSITIVE_INFINITY;

	}

        public double Train (double[] answers) throws TrainingException
        {

          double[] averageGradient;
          double[] currentDirection;
          double[] trueCurrentDirection;
          averageGradient = NeuralNet.RunNetworkAndGetAverageGradient(ol, answers);
          //averageGradient = NeuralNet.RunNetworkAndGetNetGradient(ol, ef, answers);



          //// Now Doing Conjugate Gradient ////



          beta *= ArrayUtilities.MultiplyVectors(averageGradient, (ArrayUtilities.SubtractVectors(averageGradient,previousGradient))); //(Polak-Riviere)

          //beta *= ArrayUtilities.MultiplyVectors(averageGradient, averageGradient);//the top part of our beta equation(non-polak)

          //System.err.println(beta);

          currentDirection = ArrayUtilities.AddVectors(ArrayUtilities.MultiplyArray(-1, averageGradient),ArrayUtilities.MultiplyArray(beta, previousDirection));
          trueCurrentDirection = (double[])currentDirection.clone();
          currentDirection = ArrayUtilities.CreateUnitVector(currentDirection); //will standardize the length vector

          //Utilities.PrintDoubleArray(System.err, currentDirection);



          double lastError;
          double currentError = NeuralNet.RunNetworkWithAverageError(ol, answers);






          double stepsize = 0.0325;


          double a = 0.0;
          double b = 0.0;
          double c = 0.0;
          double d = 0.0;

                  //Now test network with states based on current state + k [averageGradient] and find ideal k

                  for (;;) //we'll go as low as we can (hopefully we'll stop)
                  {
                          lastError = currentError;
                          currentError = NeuralNet.RunWithModsAndAverageError(ol, answers, stepsize, currentDirection);

                          a = b;
                          b = c;
                          c = stepsize;


                          if (currentError > lastError) //we've found our interval
                                  break;

                          stepsize *= 2.0; //otherwise double the stepsize
                  }

                  b = c;

                  //b and a are outer limits now, c and d are inner





                          c = a + goldenSection * (b - a); //set-up of golden mean

                          d = b - goldenSection * (b - a);

                          double errorC = NeuralNet.RunWithModsAndAverageError(ol, answers, c, currentDirection);
                          double errorD = NeuralNet.RunWithModsAndAverageError(ol, answers, d, currentDirection);

                          while ((b - a) > resolution)  //implementing the golden mean
                          {


                                  if ( errorC < errorD)
                                  {
                                          // a = a;
                                          b = d;
                                          d = c;
                                          c = a + goldenSection * (b - a);
                                          errorD = errorC;
                                          errorC = NeuralNet.RunWithModsAndAverageError(ol, answers, c, currentDirection);
                                  }
                                  else
                                  {
                                          a = c;
                                          //b = b;
                                          c = d;

                                          d = b - goldenSection * (b - a);

                                          errorC = errorD;
                                          errorD = NeuralNet.RunWithModsAndAverageError(ol, answers, d, currentDirection);

                                  }



                          }

                          double currError = NeuralNet.RunWithModsAndAverageError(ol, answers, (b + a) * 0.5, currentDirection);

                          if (currError > lastRanError) //reset to steepest
                          {
                                  if (beta == 0.0)
                                          throw new TrainingException("Conjugate Gradient can't train anymore");

                                  beta = 0.0;
                                  previousDirection = trueCurrentDirection;
                                  //return ArrayUtilities.SumOfSquares(NeuralNet.RunNetwork(ol, answers));
                                  //return NeuralNet.RunNetworkWithTotalError(ol, ef, answers);
                                  return lastRanError;
                          }

                          ol.ModifyNetwork(ArrayUtilities.MultiplyArray((b + a) * 0.5, currentDirection)); //only modify if we indeed found an improvement


                          //Utilities.PrintDoubleArray(System.err, ArrayUtilities.MultiplyArray((b + a) * 0.5, currentDirection));


                          //System.out.println(l);
                          if (l == n)
                          {
                                  l = 1;
                                  beta = 0.0;
                          }
                          else
                          {
                                  beta = 1.0 / ArrayUtilities.SumOfSquares(averageGradient);
                                  l++;
                          }

                  previousDirection = trueCurrentDirection;


                  previousGradient = averageGradient;
                  lastRanError = currError;

                  return currError;

	}



}

